@@ -140,7 +140,7 @@ xsltFreeAttrElem(xsltAttrElemPtr attr) {
 static void
 xsltFreeAttrElemList(xsltAttrElemPtr list) {
     xsltAttrElemPtr next;
-    
+
     while (list != NULL) {
 	next = list->next;
 	xsltFreeAttrElem(list);
@@ -171,7 +171,7 @@ xsltAddAttrElemList(xsltAttrElemPtr list, xmlNodePtr attr) {
     if (list == NULL)
 	return(xsltNewAttrElem(attr));
     cur = list;
-    while (cur != NULL) {	
+    while (cur != NULL) {
 	next = cur->next;
 	if (cur->attr == attr)
 	    return(cur);
@@ -293,13 +293,15 @@ xsltParseStylesheetAttributeSet(xsltStylesheetPtr style, xmlNodePtr cur) {
     xmlNodePtr child;
     xsltAttrElemPtr attrItems;
 
-    if ((cur == NULL) || (style == NULL))
+    if ((cur == NULL) || (style == NULL) || (cur->type != XML_ELEMENT_NODE))
 	return;
 
     value = xmlGetNsProp(cur, (const xmlChar *)"name", NULL);
-    if (value == NULL) {
+    if ((value == NULL) || (*value == 0)) {
 	xsltGenericError(xsltGenericErrorContext,
 	     "xsl:attribute-set : name is missing\n");
+        if (value)
+	    xmlFree(value);
 	return;
     }
 
@@ -375,7 +377,7 @@ xsltParseStylesheetAttributeSet(xsltStylesheetPtr style, xmlNodePtr cur) {
 			attrItems = curAttr;
 			goto next_child;
 		    }
-		    if (curAttr->next == NULL)			
+		    if (curAttr->next == NULL)
 			curAttr->next = xsltNewAttrElem(child);
 		    curAttr = nextAttr;
 		}
@@ -405,7 +407,7 @@ next_child:
     /*
     * Process attribue "use-attribute-sets".
     */
-    /* TODO check recursion */    
+    /* TODO check recursion */
     value = xmlGetNsProp(cur, (const xmlChar *)"use-attribute-sets",
 	NULL);
     if (value != NULL) {
@@ -422,7 +424,7 @@ next_child:
 		const xmlChar *ncname2 = NULL;
 		const xmlChar *prefix2 = NULL;
 		xsltAttrElemPtr refAttrItems;
-		
+
 #ifdef WITH_XSLT_DEBUG_ATTRIBUTES
 		xsltGenericDebug(xsltGenericDebugContext,
 		    "xsl:attribute-set : %s adds use %s\n", ncname, curval);
@@ -482,19 +484,27 @@ xsltGetSAS(xsltStylesheetPtr style, const xmlChar *name, const xmlChar *ns) {
 }
 
 /**
- * xsltResolveSASCallback,:
+ * xsltResolveSASCallbackInt:
  * @style:  the XSLT stylesheet
  *
  * resolve the references in an attribute set.
  */
 static void
-xsltResolveSASCallback(xsltAttrElemPtr values, xsltStylesheetPtr style,
+xsltResolveSASCallbackInt(xsltAttrElemPtr values, xsltStylesheetPtr style,
 	               const xmlChar *name, const xmlChar *ns,
-		       ATTRIBUTE_UNUSED const xmlChar *ignored) {
+		       int depth) {
     xsltAttrElemPtr tmp;
     xsltAttrElemPtr refs;
 
     tmp = values;
+    if ((name == NULL) || (name[0] == 0))
+        return;
+    if (depth > 100) {
+	xsltGenericError(xsltGenericErrorContext,
+	"xsl:attribute-set : use-attribute-sets recursion detected on %s\n",
+			 name);
+	return;
+    }
     while (tmp != NULL) {
 	if (tmp->set != NULL) {
 	    /*
@@ -519,7 +529,7 @@ xsltResolveSASCallback(xsltAttrElemPtr values, xsltStylesheetPtr style,
 		    /*
 		     * recurse first for cleanup
 		     */
-		    xsltResolveSASCallback(refs, style, name, ns, NULL);
+		    xsltResolveSASCallbackInt(refs, style, name, ns, depth + 1);
 		    /*
 		     * Then merge
 		     */
@@ -536,6 +546,19 @@ xsltResolveSASCallback(xsltAttrElemPtr values, xsltStylesheetPtr style,
     }
 }
 
+/**
+ * xsltResolveSASCallback,:
+ * @style:  the XSLT stylesheet
+ *
+ * resolve the references in an attribute set.
+ */
+static void
+xsltResolveSASCallback(xsltAttrElemPtr values, xsltStylesheetPtr style,
+	               const xmlChar *name, const xmlChar *ns,
+		       ATTRIBUTE_UNUSED const xmlChar *ignored) {
+    xsltResolveSASCallbackInt(values, style, name, ns, 1);
+}
+
 /**
  * xsltMergeSASCallback,:
  * @style:  the XSLT stylesheet
@@ -605,7 +628,7 @@ xsltResolveStylesheetAttributeSet(xsltStylesheetPtr style) {
 #endif
 		style->attributeSets = xmlHashCreate(10);
 	    }
-	    xmlHashScanFull(cur->attributeSets, 
+	    xmlHashScanFull(cur->attributeSets,
 		(xmlHashScannerFull) xsltMergeSASCallback, style);
 	    /*
 	     * the attribute lists have either been migrated to style
@@ -621,7 +644,7 @@ xsltResolveStylesheetAttributeSet(xsltStylesheetPtr style) {
      * Then resolve all the references and computes the resulting sets
      */
     if (style->attributeSets != NULL) {
-	xmlHashScanFull(style->attributeSets, 
+	xmlHashScanFull(style->attributeSets,
 		(xmlHashScannerFull) xsltResolveSASCallback, style);
     }
 }
@@ -629,7 +652,7 @@ xsltResolveStylesheetAttributeSet(xsltStylesheetPtr style) {
 /**
  * xsltAttributeInternal:
  * @ctxt:  a XSLT process context
- * @node:  the node in the source tree.
+ * @node:  the current node in the source tree
  * @inst:  the xsl:attribute element
  * @comp:  precomputed information
  * @fromAttributeSet:  the attribute comes from an attribute-set
@@ -637,31 +660,34 @@ xsltResolveStylesheetAttributeSet(xsltStylesheetPtr style) {
  * Process the xslt attribute node on the source node
  */
 static void
-xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
+xsltAttributeInternal(xsltTransformContextPtr ctxt,
+		      xmlNodePtr contextNode,
                       xmlNodePtr inst,
 		      xsltStylePreCompPtr castedComp,
-                      int fromAttributeSet) {
+                      int fromAttributeSet)
+{
 #ifdef XSLT_REFACTORED
     xsltStyleItemAttributePtr comp =
 	(xsltStyleItemAttributePtr) castedComp;
-    /*
-    * TODO: Change the fields of the compiled struct:
-    *  1) @name (char)
-    *  2) @nameType (String, AVT)
-    *  3) @nsName (char)
-    *  4) nsNameType (None, String, AVT)
-    */
 #else
     xsltStylePreCompPtr comp = castedComp;
 #endif
     xmlNodePtr targetElem;
-    xmlChar *prop = NULL;    
+    xmlChar *prop = NULL;
     const xmlChar *name = NULL, *prefix = NULL, *nsName = NULL;
     xmlChar *value = NULL;
     xmlNsPtr ns = NULL;
-    xmlAttrPtr attr;    
+    xmlAttrPtr attr;
 
-    if ((ctxt == NULL) || (node == NULL) || (inst == NULL))
+    if ((ctxt == NULL) || (contextNode == NULL) || (inst == NULL) ||
+        (inst->type != XML_ELEMENT_NODE) )
+        return;
+
+    /*
+    * A comp->has_name == 0 indicates that we need to skip this instruction,
+    * since it was evaluated to be invalid already during compilation.
+    */
+    if (!comp->has_name)
         return;
     /*
     * BIG NOTE: This previously used xsltGetSpecialNamespace() and
@@ -679,15 +705,15 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
     if (comp == NULL) {
         xsltTransformError(ctxt, NULL, inst,
 	    "Internal error in xsltAttributeInternal(): "
-	    "The instruction was no compiled.\n");
+	    "The XSLT 'attribute' instruction was not compiled.\n");
         return;
-    }    
+    }
     /*
     * TODO: Shouldn't ctxt->insert == NULL be treated as an internal error?
     *   So report an internal error?
     */
     if (ctxt->insert == NULL)
-        return;    
+        return;
     /*
     * SPEC XSLT 1.0:
     *  "Adding an attribute to a node that is not an element;
@@ -699,7 +725,7 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
     targetElem = ctxt->insert;
     if (targetElem->type != XML_ELEMENT_NODE)
 	return;
-    
+
     /*
     * SPEC XSLT 1.0:
     * "Adding an attribute to an element after children have been added
@@ -726,12 +752,10 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
     * Process the name
     * ----------------
     */
-    if (!comp->has_name) /* TODO: raise error */
-        return;
 
 #ifdef WITH_DEBUGGER
     if (ctxt->debugStatus != XSLT_DEBUG_NONE)
-        xslHandleDebugger(inst, node, NULL, ctxt);
+        xslHandleDebugger(inst, contextNode, NULL, ctxt);
 #endif
 
     if (comp->name == NULL) {
@@ -749,36 +773,45 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
 		"valid QName.\n", prop);
 	    /* we fall through to catch any further errors, if possible */
 	}
+
+	/*
+	* Reject a name of "xmlns".
+	*/
+	if (xmlStrEqual(prop, BAD_CAST "xmlns")) {
+            xsltTransformError(ctxt, NULL, inst,
+                "xsl:attribute: The effective name 'xmlns' is not allowed.\n");
+	    xmlFree(prop);
+	    goto error;
+	}
+
 	name = xsltSplitQName(ctxt->dict, prop, &prefix);
 	xmlFree(prop);
     } else {
+	/*
+	* The "name" value was static.
+	*/
+#ifdef XSLT_REFACTORED
+	prefix = comp->nsPrefix;
+	name = comp->name;
+#else
 	name = xsltSplitQName(ctxt->dict, comp->name, &prefix);
+#endif
     }
 
-    if (!xmlStrncasecmp(prefix, (xmlChar *) "xmlns", 5)) {
-#ifdef WITH_XSLT_DEBUG_PARSING
-        xsltGenericDebug(xsltGenericDebugContext,
-	    "xsltAttribute: xmlns prefix forbidden\n");
-#endif
-	/*
-	* SPEC XSLT 1.0:
-	*  "It is an error if the string that results from instantiating
-	*  the attribute value template is not a QName or is the string
-	*  xmlns. An XSLT processor may signal the error; if it does not
-	*  signal the error, it must recover by not adding the attribute
-	*  to the result tree."
-	* TODO: Decide which way to go here.
-	*/
-        goto error;
-    }
     /*
     * Process namespace semantics
     * ---------------------------
     *
     * Evaluate the namespace name.
     */
-    if (comp->has_ns) {	 
+    if (comp->has_ns) {
+	/*
+	* The "namespace" attribute was existent.
+	*/
 	if (comp->ns != NULL) {
+	    /*
+	    * No AVT; just plain text for the namespace name.
+	    */
 	    if (comp->ns[0] != 0)
 		nsName = comp->ns;
 	} else {
@@ -788,9 +821,9 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	    */
 	    /* TODO: check attr acquisition wrt to the XSLT namespace */
 	    tmpNsName = xsltEvalAttrValueTemplate(ctxt, inst,
-		(const xmlChar *) "namespace", XSLT_NAMESPACE);	
+		(const xmlChar *) "namespace", XSLT_NAMESPACE);
 	    /*
-	    * This fixes bug #302020: The AVT might also evaluate to the 
+	    * This fixes bug #302020: The AVT might also evaluate to the
 	    * empty string; this means that the empty string also indicates
 	    * "no namespace".
 	    * SPEC XSLT 1.0:
@@ -799,8 +832,20 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	    */
 	    if ((tmpNsName != NULL) && (tmpNsName[0] != 0))
 		nsName = xmlDictLookup(ctxt->dict, BAD_CAST tmpNsName, -1);
-	    xmlFree(tmpNsName);		
-	};	    
+	    xmlFree(tmpNsName);
+	}
+
+        if (xmlStrEqual(nsName, BAD_CAST "http://www.w3.org/2000/xmlns/")) {
+            xsltTransformError(ctxt, NULL, inst,
+                "xsl:attribute: Namespace http://www.w3.org/2000/xmlns/ "
+                "forbidden.\n");
+            goto error;
+        }
+        if (xmlStrEqual(nsName, XML_XML_NAMESPACE)) {
+            prefix = BAD_CAST "xml";
+        } else if (xmlStrEqual(prefix, BAD_CAST "xml")) {
+            prefix = NULL;
+        }
     } else if (prefix != NULL) {
 	/*
 	* SPEC XSLT 1.0:
@@ -808,7 +853,7 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	*  expanded into an expanded-name using the namespace declarations
 	*  in effect for the xsl:attribute element, *not* including any
 	*  default namespace declaration."
-	*/	
+	*/
 	ns = xmlSearchNs(inst->doc, inst, prefix);
 	if (ns == NULL) {
 	    /*
@@ -816,42 +861,37 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	    *  Saxon, Xalan-J and MSXML).
 	    */
 	    xsltTransformError(ctxt, NULL, inst,
-		"xsl:attribute: The effective prefix '%s', has no "
+		"xsl:attribute: The QName '%s:%s' has no "
 		"namespace binding in scope in the stylesheet; "
 		"this is an error, since the namespace was not "
-		"specified by the instruction itself.\n", prefix);
+		"specified by the instruction itself.\n", prefix, name);
 	} else
-	    nsName = ns->href;	
+	    nsName = ns->href;
     }
 
     if (fromAttributeSet) {
 	/*
-	* I think this tries to ensure that xsl:attribute(s) coming
+	* This tries to ensure that xsl:attribute(s) coming
 	* from an xsl:attribute-set won't override attribute of
 	* literal result elements or of explicit xsl:attribute(s).
+	* URGENT TODO: This might be buggy, since it will miss to
+	*  overwrite two equal attributes both from attribute sets.
 	*/
 	attr = xmlHasNsProp(targetElem, name, nsName);
 	if (attr != NULL)
 	    return;
     }
 
-    /*
-    * Something about ns-prefixes:
-    * SPEC XSLT 1.0:
-    *  "XSLT processors may make use of the prefix of the QName specified
-    *  in the name attribute when selecting the prefix used for outputting
-    *  the created attribute as XML; however, they are not required to do
-    *  so and, if the prefix is xmlns, they must not do so"
-    */        
     /*
     * Find/create a matching ns-decl in the result tree.
     */
     ns = NULL;
+
 #if 0
-    if (0) {	
+    if (0) {
 	/*
 	* OPTIMIZE TODO: How do we know if we are adding to a
-	*  fragment or not?
+	*  fragment or to the result tree?
 	*
 	* If we are adding to a result tree fragment (i.e., not to the
 	* actual result tree), we'll don't bother searching for the
@@ -859,100 +899,43 @@ xsltAttributeInternal(xsltTransformContextPtr ctxt, xmlNodePtr node,
 	* tree fragment.
 	*/
 	if (nsName != NULL) {
-	    ns = xsltTreeAcquireStoredNs(ctxt->document->doc, nsName,
-		prefix);
+	    /*
+	    * TODO: Get the doc of @targetElem.
+	    */
+	    ns = xsltTreeAcquireStoredNs(some doc, nsName, prefix);
 	}
     }
 #endif
+
     if (nsName != NULL) {
 	/*
-	* The owner-element might be in the same namespace.
+	* Something about ns-prefixes:
+	* SPEC XSLT 1.0:
+	*  "XSLT processors may make use of the prefix of the QName specified
+	*  in the name attribute when selecting the prefix used for outputting
+	*  the created attribute as XML; however, they are not required to do
+	*  so and, if the prefix is xmlns, they must not do so"
 	*/
-	if ((targetElem->ns != NULL) &&
-	    (targetElem->ns->prefix != NULL) &&	    
-	    xmlStrEqual(targetElem->ns->href, nsName))
-	{
-	    ns = targetElem->ns;
-	    goto namespace_finished;
-	}
-	if (prefix != NULL) {
-	    /*
-	    * Search by ns-prefix.
-	    */
-	    ns = xmlSearchNs(targetElem->doc, targetElem, prefix);
-	    if ((ns != NULL) && (xmlStrEqual(ns->href, nsName))) {
-		goto namespace_finished;
-	    }
-	}
 	/*
-	* Fallback to a search by ns-name.
-	*/	
-	ns = xmlSearchNsByHref(targetElem->doc, targetElem, nsName);
-	if ((ns != NULL) && (ns->prefix != NULL)) {
-	    goto namespace_finished;
-	}
-	/*
-	* OK, we need to declare the namespace on the target element.
+	* xsl:attribute can produce a scenario where the prefix is NULL,
+	* so generate a prefix.
 	*/
-	if (prefix) {
-	    if (targetElem->nsDef != NULL) {
-		ns = targetElem->nsDef;
-		do {
-		    if ((ns->prefix) && xmlStrEqual(ns->prefix, prefix)) {
-			/*
-			* The prefix is aready occupied.
-			*/
-			break;
-		    }
-		    ns = ns->next;
-		} while (ns != NULL);
-		if (ns == NULL) {
-		    ns = xmlNewNs(targetElem, nsName, prefix);
-		    goto namespace_finished;
-		}
-	    }
-	}
-	/*
-	* Generate a new prefix.
-	*/
-	{
-	    const xmlChar *basepref = prefix;
-	    xmlChar pref[30];
-	    int counter = 1;
-	    
-	    if (prefix != NULL)
-		basepref = prefix;
-	    else
-		basepref = xmlStrdup(BAD_CAST "ns");
-	    
-	    do {
-		snprintf((char *) pref, 30, "%s%d",
-		    basepref, counter++);
-		ns = xmlSearchNs(targetElem->doc, targetElem, BAD_CAST pref);
-		if (counter > 1000) {
-		    xsltTransformError(ctxt, NULL, inst,
-		    	"Namespace fixup error: Failed to compute a "
-			"new unique ns-prefix for the generated attribute "
-			"{%s}%s'.\n", nsName, name);		    			
-		    ns = NULL;
-		    break;
-		}
-	    } while (ns != NULL);
-	    if (basepref != prefix)
-		xmlFree((xmlChar *)basepref);
-	    ns = xmlNewNs(targetElem, nsName, BAD_CAST pref);
-	}
+	if ((prefix == NULL) || xmlStrEqual(prefix, BAD_CAST "xmlns")) {
+	    xmlChar *pref = xmlStrdup(BAD_CAST "ns_1");
 
-namespace_finished:
+	    ns = xsltGetSpecialNamespace(ctxt, inst, nsName, pref, targetElem);
 
+	    xmlFree(pref);
+	} else {
+	    ns = xsltGetSpecialNamespace(ctxt, inst, nsName, prefix,
+		targetElem);
+	}
 	if (ns == NULL) {
 	    xsltTransformError(ctxt, NULL, inst,
 		"Namespace fixup error: Failed to acquire an in-scope "
-		"namespace binding of the generated attribute '{%s}%s'.\n",
+		"namespace binding for the generated attribute '{%s}%s'.\n",
 		nsName, name);
-	    /*
-	    * TODO: Should we just stop here?
-	    */
+	    goto error;
 	}
     }
     /*
@@ -965,15 +948,14 @@ namespace_finished:
 	* TODO: Do we need to put the empty string in ?
 	*/
 	attr = xmlSetNsProp(ctxt->insert, ns, name, (const xmlChar *) "");
-    } else if ((inst->children->next == NULL) && 
+    } else if ((inst->children->next == NULL) &&
 	    ((inst->children->type == XML_TEXT_NODE) ||
 	     (inst->children->type == XML_CDATA_SECTION_NODE)))
     {
-	xmlNodePtr origTxt = inst->children, copyTxt;
+	xmlNodePtr copyTxt;
+
 	/*
-	* Optimization: if the content is just 1 text node, then
-	* just need to copy it over, or just assign it to the result
-	* if the string is shared.
+	* xmlSetNsProp() will take care of duplicates.
 	*/
 	attr = xmlSetNsProp(ctxt->insert, ns, name, NULL);
 	if (attr == NULL) /* TODO: report error ? */
@@ -992,44 +974,53 @@ namespace_finished:
 	    * This is a safe scenario where we don't need to lookup
 	    * the dict.
 	    */
-	    copyTxt->content = origTxt->content;
+	    copyTxt->content = inst->children->content;
 	    /*
 	    * Copy "disable-output-escaping" information.
 	    * TODO: Does this have any effect for attribute values
 	    *  anyway?
 	    */
-	    if (origTxt->name == xmlStringTextNoenc)
+	    if (inst->children->name == xmlStringTextNoenc)
 		copyTxt->name = xmlStringTextNoenc;
 	} else {
 	    /*
 	    * Copy the value.
 	    */
-	    copyTxt = xmlNewText(origTxt->content);
+	    copyTxt = xmlNewText(inst->children->content);
 	    if (copyTxt == NULL) /* TODO: report error */
 		goto error;
-	    /*
-	    * Copy "disable-output-escaping" information.
-	    * TODO: Does this have any effect for attribute values
-	    *  anyway?
-	    */
-	    if (origTxt->name == xmlStringTextNoenc)
-		copyTxt->name = xmlStringTextNoenc;
-	}
-	if (copyTxt != NULL) {
-	    copyTxt->doc = attr->doc;
-	    xmlAddChild((xmlNodePtr) attr, copyTxt);
 	}
+	attr->children = attr->last = copyTxt;
+	copyTxt->parent = (xmlNodePtr) attr;
+	copyTxt->doc = attr->doc;
+	/*
+	* Copy "disable-output-escaping" information.
+	* TODO: Does this have any effect for attribute values
+	*  anyway?
+	*/
+	if (inst->children->name == xmlStringTextNoenc)
+	    copyTxt->name = xmlStringTextNoenc;
+
+        /*
+         * since we create the attribute without content IDness must be
+         * asserted as a second step
+         */
+        if ((copyTxt->content != NULL) &&
+            (xmlIsID(attr->doc, attr->parent, attr)))
+            xmlAddID(NULL, attr->doc, copyTxt->content, attr);
     } else {
 	/*
 	* The sequence constructor might be complex, so instantiate it.
 	*/
-	value = xsltEvalTemplateString(ctxt, node, inst);
+	value = xsltEvalTemplateString(ctxt, contextNode, inst);
 	if (value != NULL) {
 	    attr = xmlSetNsProp(ctxt->insert, ns, name, value);
 	    xmlFree(value);
 	} else {
 	    /*
 	    * TODO: Do we have to add the empty string to the attr?
+	    * TODO: Does a  value of NULL indicate an
+	    *  error in xsltEvalTemplateString() ?
 	    */
 	    attr = xmlSetNsProp(ctxt->insert, ns, name,
 		(const xmlChar *) "");
@@ -1037,7 +1028,7 @@ namespace_finished:
     }
 
 error:
-    return;    
+    return;
 }
 
 /**
@@ -1059,69 +1050,100 @@ xsltAttribute(xsltTransformContextPtr ctxt, xmlNodePtr node,
  * xsltApplyAttributeSet:
  * @ctxt:  the XSLT stylesheet
  * @node:  the node in the source tree.
- * @inst:  the xslt attribute node
- * @attributes:  the set list.
+ * @inst:  the attribute node "xsl:use-attribute-sets"
+ * @attrSets:  the list of QNames of the attribute-sets to be applied
  *
- * Apply the xsl:use-attribute-sets
+ * Apply the xsl:use-attribute-sets.
+ * If @attrSets is NULL, then @inst will be used to exctract this
+ * value.
+ * If both, @attrSets and @inst, are NULL, then this will do nothing.
  */
-
 void
 xsltApplyAttributeSet(xsltTransformContextPtr ctxt, xmlNodePtr node,
-                      xmlNodePtr inst ATTRIBUTE_UNUSED,
-                      const xmlChar * attributes)
+                      xmlNodePtr inst,
+                      const xmlChar *attrSets)
 {
     const xmlChar *ncname = NULL;
     const xmlChar *prefix = NULL;
-    const xmlChar *attrib, *endattr;
-    xsltAttrElemPtr values;
+    const xmlChar *curstr, *endstr;
+    xsltAttrElemPtr attrs;
     xsltStylesheetPtr style;
 
-    if (attributes == NULL) {
-        return;
-    }
+    if (attrSets == NULL) {
+	if (inst == NULL)
+	    return;
+	else {
+	    /*
+	    * Extract the value from @inst.
+	    */
+	    if (inst->type == XML_ATTRIBUTE_NODE) {
+		if ( ((xmlAttrPtr) inst)->children != NULL)
+		    attrSets = ((xmlAttrPtr) inst)->children->content;
 
-    attrib = attributes;
-    while (*attrib != 0) {
-        while (IS_BLANK(*attrib))
-            attrib++;
-        if (*attrib == 0)
+	    }
+	    if (attrSets == NULL) {
+		/*
+		* TODO: Return an error?
+		*/
+		return;
+	    }
+	}
+    }
+    /*
+    * Parse/apply the list of QNames.
+    */
+    curstr = attrSets;
+    while (*curstr != 0) {
+        while (IS_BLANK(*curstr))
+            curstr++;
+        if (*curstr == 0)
             break;
-        endattr = attrib;
-        while ((*endattr != 0) && (!IS_BLANK(*endattr)))
-            endattr++;
-        attrib = xmlDictLookup(ctxt->dict, attrib, endattr - attrib);
-        if (attrib) {
-#ifdef WITH_XSLT_DEBUG_ATTRIBUTES
+        endstr = curstr;
+        while ((*endstr != 0) && (!IS_BLANK(*endstr)))
+            endstr++;
+        curstr = xmlDictLookup(ctxt->dict, curstr, endstr - curstr);
+        if (curstr) {
+	    /*
+	    * TODO: Validate the QName.
+	    */
+
+#ifdef WITH_XSLT_DEBUG_curstrUTES
             xsltGenericDebug(xsltGenericDebugContext,
-                             "apply attribute set %s\n", attrib);
+                             "apply curstrute set %s\n", curstr);
 #endif
-            ncname = xsltSplitQName(ctxt->dict, attrib, &prefix);
+            ncname = xsltSplitQName(ctxt->dict, curstr, &prefix);
 
             style = ctxt->style;
+
 #ifdef WITH_DEBUGGER
-            if ((style != NULL) && (style->attributeSets != NULL) &&
-		(ctxt->debugStatus != XSLT_DEBUG_NONE)) {
-                values =
+            if ((style != NULL) &&
+		(style->attributeSets != NULL) &&
+		(ctxt->debugStatus != XSLT_DEBUG_NONE))
+	    {
+                attrs =
                     xmlHashLookup2(style->attributeSets, ncname, prefix);
-                if ((values != NULL) && (values->attr != NULL))
-                    xslHandleDebugger(values->attr->parent, node, NULL,
-                                      ctxt);
+                if ((attrs != NULL) && (attrs->attr != NULL))
+                    xslHandleDebugger(attrs->attr->parent, node, NULL,
+			ctxt);
             }
 #endif
+	    /*
+	    * Lookup the referenced curstrute-set.
+	    */
             while (style != NULL) {
-                values =
+                attrs =
                     xmlHashLookup2(style->attributeSets, ncname, prefix);
-                while (values != NULL) {
-                    if (values->attr != NULL) {
-                        xsltAttributeInternal(ctxt, node, values->attr,
-                                              values->attr->psvi, 1);
+                while (attrs != NULL) {
+                    if (attrs->attr != NULL) {
+                        xsltAttributeInternal(ctxt, node, attrs->attr,
+			    attrs->attr->psvi, 1);
                     }
-                    values = values->next;
+                    attrs = attrs->next;
                 }
                 style = xsltNextImport(style);
             }
         }
-        attrib = endattr;
+        curstr = endstr;
     }
 }
 
