@@ -1639,6 +1639,8 @@ struct _xsltStylesheet {
      * Forwards-compatible processing
      */
     int forwards_compatible;
+
+    xmlHashTablePtr namedTemplates; /* hash table of named templates */
 };
 
 typedef struct _xsltTransformCache xsltTransformCache;
@@ -1909,14 +1911,6 @@ XSLTPUBFUN void XSLTCALL
 XSLTPUBFUN void XSLTCALL
 			xsltReleaseRVT		(xsltTransformContextPtr ctxt,
 						 xmlDocPtr RVT);
-XSLTPUBFUN int XSLTCALL
-			xsltTransStorageAdd	(xsltTransformContextPtr ctxt,
-						 void *id,
-						 void *data);
-XSLTPUBFUN void * XSLTCALL
-			xsltTransStorageRemove	(xsltTransformContextPtr ctxt,
-						 void *id);
-
 /*
  * Extra functions for Attribute Value Templates
  */
